@@ -8,6 +8,7 @@
 """
 
 import os
+import pprint
 import re
 import sys
 import unicodedata
@@ -17,25 +18,25 @@ import sphinx
 from docutils.parsers.rst import Directive
 from docutils.transforms import Transform
 
-SUPORTED_NODETYPES = [
+SUPORTED_NODETYPES_POSITIONS = {
 #    'document',
-    'section',
-    'title',
-    'paragraph',
-    'block_quote',
-    'literal_block',
-    'note',
-    'tip',
-    'warning',
-    'important',
-    'bullet_list',
+    'section': ['before', 'after', 'inside'],
+    'title': ['after'],
+    'paragraph': ['before', 'after'],
+    'block_quote': ['before', 'after'],
+    'literal_block': ['before', 'after'],
+    'note': ['before', 'after'],
+    'tip': ['before', 'after'],
+    'warning': ['before', 'after'],
+    'important': ['before', 'after'],
+    'bullet_list': ['before', 'after', 'inside'],
 #    'list_item',
-    'figure',
+    'figure': ['before', 'after'],
 #    'caption',
-    'toctree',
+    'toctree': ['before', 'after', 'inside'],
 #    'compound',
-    'comment',
-    ]
+    'comment': ['before', 'after'],
+    }
 
 existing_ids = set()
 inherits = {}
@@ -104,9 +105,93 @@ def generate_inheritref(node):
     return identifier
 
 
+class InheritRef(Directive):
+    has_content = True
+    #required_arguments = 0
+    #optional_arguments = 1
+    #final_argument_witespace = False
+    #option_spec = {
+    #    'hola': directives.unchanged_required,
+    #    }
+
+    def run(self):
+        #env = self.state.document.settings.env
+        config = self.state.document.settings.env.config
+        reference_pattern = config.inheritance_reference_pattern
+        if isinstance(reference_pattern, basestring):
+            reference_pattern = re.compile(reference_pattern)
+
+        if not self.content or not self.content[0]:
+            raise Exception("ERROR: 'inheritref' directive requires content: "
+                    "'%s'" % self.content)
+        ref = self.content[0]
+        ref_match = reference_pattern.match(ref)
+        if not ref_match:
+            raise Exception("ERROR: invalid inheritref '%s'. It doesn't have "
+                    "the expected format: PREFIX:NODE_TYPE:IDENTIFIER" % ref)
+        nodetype = ref_match.groupdict().get('nodetype', '')
+        if nodetype not in SUPORTED_NODETYPES_POSITIONS:
+            raise Exception("ERROR: invalid inheritref '%s'. The node type "
+                    "'%s' is not supported" % (ref, nodetype))
+        if ref in existing_ids:
+            raise Exception("ERROR: explicit inheritref '%s' is defined more "
+                    "than one times." % ref)
+        existing_ids.add(ref)
+        if len(self.content) > 1:
+            sys.stderr.write("WARNING: Unexpected extra content in directive "
+                    "'inheritref' with ref '%s'\n" % ref)
+        inheritref = inheritref_node('')
+        inheritref['inheritref'] = ref
+        inheritref['inheritnodetype'] = nodetype
+        if config.verbose and nodetype == 'toctree':
+            source = (self.state.document.attributes['source'] or '')
+            sys.stderr.write("DEBUG: toctree node %s in file %s. Exists in "
+                    "inheritance dict? %s.\n" % (ref, source, ref in inherits))
+        return [inheritref]
+
+
 class Replacer(Transform):
     default_priority = 1000
 
+    @classmethod
+    def _search_inherited_node(cls, curr_node, config):
+        assert isinstance(curr_node, inheritref_node), "Unexpected type of " \
+                "curr_node: %s (%s). cls: %s" % (curr_node, type(curr_node), cls)
+        ref = curr_node['inheritref']
+        inherited_node_type = curr_node['inheritnodetype']
+        node_index = curr_node.parent.index(curr_node)
+        # Get next node in doctree. If 'inheritref' node is the last in current
+        # "parent" (typically; section) it search in next parent brother
+        while (curr_node.parent and
+                (len(curr_node.parent) - 1) == node_index):
+            curr_node = curr_node.parent
+            node_index = curr_node.parent.index(curr_node)
+        if not curr_node.parent:
+            raise Exception("ERROR: inheritref node '%s' doesn't "
+                    "have any next node." % ref)
+        next_node = curr_node.parent[node_index + 1]
+        if config.verbose:
+            sys.stderr.write("DEBUG: next_node: %s. Now, I'll try to get "
+                    "traverse:\n" % next_node)
+        for tnode in next_node.traverse():
+            if tnode.tagname == inherited_node_type:
+                if config.verbose:
+                    sys.stderr.write("  - %s. This is the inherited node\n"
+                            % tnode)
+                return tnode
+            if config.verbose:
+                sys.stderr.write("  - %s.\n" % tnode)
+            if tnode.tagname == 'inheritref_node':
+                raise Exception("Found unexpected inheritref node before found "
+                        "the inherited node of previous inheritref: %s" % tnode)
+        if config.verbose:
+            sys.stderr.write("  END traverse next_node.\n")
+        if next_node.tagname == inherited_node_type:
+            return next_node
+        raise Exception("Invalid inheritref definition '%s'. The type of "
+                "inherited node found is not the expected: %s"
+                        % (ref, next_node.tagname))
+
     def apply(self):
         config = self.document.settings.env.config
         pattern = config.inheritance_pattern
@@ -114,9 +199,25 @@ class Replacer(Transform):
             pattern = re.compile(pattern)
 
         current_inherit_ref = None
+        current_inherit_vals = None
         for node in self.document.traverse():
             if isinstance(node, (docutils.nodes.Inline, docutils.nodes.Text)):
                 continue
+            if isinstance(node, inheritref_node):
+                inherit_node = Replacer._search_inherited_node(node, config)
+                inherit_node_type = node['inheritnodetype']
+                node['inheritnode'] = inherit_node
+                if (inherit_node_type not in ('section', 'title') or
+                        node.parent.index(node) <= 1):
+                    continue
+                # inheritref node introduced by inheritref directive is
+                # introduced in a wrong position when nodetype is 'section' or
+                # title; it's introduced as child of previous section when it
+                # as previous simbling node respect inherited node
+                node.parent.replace(node, [])
+                inherit_node.replace_self([node, inherit_node])
+                continue
+
             parent = node.parent
             text = node.astext()
             match = pattern.search(text)
@@ -129,20 +230,34 @@ class Replacer(Transform):
                         'inherit_issue_pattern must have '
                         'exactly one group: {0!r}'.format(match.groups()))
 
+                if current_inherit_ref is not None:
+                    self._check_inherit_vals(current_inherit_ref,
+                            current_inherit_vals)
+
                 # extract the reference data (excluding the leading dash)
                 refdata = match.group(1)
 
                 source = (node.document and node.document.attributes['source'] 
                     or '')
                 try:
-                    position, refsource, reftype, refid = \
+                    position, refsource, nodetype, refid = \
                             refdata.split(':')
                 except ValueError:
-                    raise ValueError('Invalid inheritance ref "%s" at %s:%s' % (
-                            refdata, source, node.line))
-                ref = '%s:%s:%s' % (refsource, reftype, refid)
+                    raise ValueError('Invalid inheritance ref "%s" at %s:%s'
+                            % (refdata, source, node.line))
+                if nodetype not in SUPORTED_NODETYPES_POSITIONS:
+                    raise Exception('ERROR: Invalid inheritance ref "%s" at '
+                            '%s:%s. The node type "%s" is not supported'
+                                    % (refdata, source, node.line, nodetype))
+                if position not in SUPORTED_NODETYPES_POSITIONS[nodetype]:
+                    raise Exception('ERROR: Invalid inheritance ref "%s" at '
+                            '%s:%s. The position "%s" is not supported for '
+                            'node types "%s"' % (refdata, source, node.line,
+                                position, nodetype))
+                ref = '%s:%s:%s' % (refsource, nodetype, refid)
                 current_inherit_ref = ref
                 current_inherit_vals = {
+                    'nodetype': nodetype,
                     'position': position,
                     'nodes': [],
                     'replaced': 0,
@@ -169,51 +284,44 @@ class Replacer(Transform):
                     if found:
                         continue
                 current_inherit_vals['nodes'].append(node)
+        if current_inherit_ref is not None:
+            self._check_inherit_vals(current_inherit_ref, current_inherit_vals)
 
+    def _check_inherit_vals(self, inherit_ref, inherit_vals):
+        config = self.document.settings.env.config
+        def get_and_check_size_type(position, nodetype, node_list, min_size,
+                max_size, nodetype_1st_node):
+            assert ((min_size is None or len(node_list) >= min_size) and
+                    (max_size is None or len(node_list) <= max_size)), \
+                    "Unexpected inheritance nodes list size in '%s' '%s' " \
+                    "inheritance '%s': %s\n    %s\n" % (position, nodetype,
+                            inherit_ref, len(node_list), node_list)
+            if len(node_list) < 1:
+                return None
+            assert isinstance(node_list[0], nodetype_1st_node), \
+                    "Unexpected type of inheritance node in '%s' '%s' " \
+                    "inheritance (%s): %s\n" % (position, nodetype,
+                            inherit_ref, type(node_list))
+            return node_list[0]
 
-class InheritRef(Directive):
-    has_content = True
-    #required_arguments = 0
-    #optional_arguments = 1
-    #final_argument_witespace = False
-    #option_spec = {
-    #    'hola': directives.unchanged_reuired,
-    #    }
-
-    def run(self):
-        #env = self.state.document.settings.env
-        config = self.state.document.settings.env.config
-        reference_pattern = config.inheritance_reference_pattern
-        if isinstance(reference_pattern, basestring):
-            reference_pattern = re.compile(reference_pattern)
-
-        if not self.content or not self.content[0]:
-            raise Exception("ERROR: 'inheritref' directive requires content: "
-                    "'%s'" % self.content)
-        ref = self.content[0]
-        ref_match = reference_pattern.match(ref)
-        if not ref_match:
-            raise Exception("ERROR: invalid inheritref '%s'. It doesn't have "
-                    "the expected format: PREFIX:NODE_TYPE:IDENTIFIER" % ref)
-        nodetype = ref_match.groupdict().get('nodetype', '')
-        if nodetype not in SUPORTED_NODETYPES:
-            raise Exception("ERROR: invalid inheritref '%s'. The node type "
-                    "'%s' is not supported" % (ref, nodetype))
-        if ref in existing_ids:
-            raise Exception("ERROR: explicit inheritref '%s' is defined more "
-                    "than one times." % ref)
-        existing_ids.add(ref)
-        if len(self.content) > 1:
-            sys.stderr.write("WARNING: Unexpected extra content in directive "
-                    "'inheritref' with ref '%s'\n" % ref)
-        inheritref = inheritref_node('')
-        inheritref['inheritref'] = ref
-        inheritref['inheritnodetype'] = nodetype
-        if config.verbose and nodetype == 'toctree':
-            source = (self.state.document.attributes['source'] or '')
-            sys.stderr.write("DEBUG: toctree node %s in file %s. Exists in "
-                    "inheritance dict? %s.\n" % (ref, source, ref in inherits))
-        return [inheritref]
+        if (inherit_vals['position'] == 'inside' and
+                inherit_vals['nodetype'] == 'toctree'):
+            inh_node = get_and_check_size_type('inside', 'toctree',
+                    inherit_vals['nodes'], 1, 1, docutils.nodes.compound)
+            inh_node = get_and_check_size_type('inside', 'toctree',
+                    inh_node.children, 1, 1, sphinx.addnodes.toctree)
+            inherit_vals['nodes'] = [inh_node]
+        elif (inherit_vals['position'] == 'inside' and
+                inherit_vals['nodetype'] == 'bullet_list'):
+            if config.verbose:
+                sys.stderr.write("inherit_vals bullet_list: %s\n" %
+                        pprint.pformat(inherit_vals, indent=2))
+            inh_node = get_and_check_size_type('inside', 'bullet_list',
+                    inherit_vals['nodes'], 1, None, docutils.nodes.bullet_list)
+            if config.verbose:
+                sys.stderr.write("inh_node bullet_list: %s\nchildren: %s\n"
+                        % (inh_node, inh_node.children))
+        return
 
 
 def init_transformer(app):
@@ -237,55 +345,19 @@ def check_module(app, docname, text):
         text[0] = ''
 
 
-def __search_inherited_node(curr_node, config):
-    assert isinstance(curr_node, inheritref_node), "Unexpected type of " \
-            "curr_node: %s." % type(curr_node)
-    ref = curr_node['inheritref']
-    inherited_node_type = curr_node['inheritnodetype']
-    if ref not in inherits:
-        if config.inheritance_debug:
-            sys.stderr.write("DEBUG: ref '%s' found in inheritref directive "
-                    "doesn't have any inheritance.\n" % ref)
-        return None
-    node_index = curr_node.parent.index(curr_node)
-    # Get next node in doctree. If 'inheritref' node is the last in current
-    # "parent" (typically; section) it search in next parent brother
-    while (curr_node.parent and
-            (len(curr_node.parent) - 1) == node_index):
-        curr_node = curr_node.parent
-        node_index = curr_node.parent.index(curr_node)
-    if not curr_node.parent:
-        raise Exception("ERROR: inheritref node '%s' doesn't "
-                "have any next node." % ref)
-    next_node = curr_node.parent[node_index + 1]
-    if config.verbose:
-        sys.stderr.write("DEBUG: next_node: %s. Now, I'll try to get "
-                "traverse:\n" % next_node)
-    for tnode in next_node.traverse():
-        if tnode.tagname == inherited_node_type:
-            if config.verbose:
-                sys.stderr.write("  - %s. This is the inherited node\n"
-                        % tnode)
-            return tnode
-        if config.verbose:
-            sys.stderr.write("  - %s.\n" % tnode)
-        if tnode.tagname == 'inheritref_node':
-            raise Exception("Found unexpected inheritref node before found "
-                    "the inherited node of previous inheritref: %s" % tnode)
-    if config.verbose:
-        sys.stderr.write("  END traverse next_node.\n")
-    return next_node
-
-
 def search_inheritances(app, doctree):
     if not doctree:
         return
     if not app.config.inheritance_autoreferences:
         for node in doctree.traverse(inheritref_node):
-            inherit_node = __search_inherited_node(node, app.config)
-            if inherit_node:
-                apply_inheritance(app, [node, inherit_node],
-                        node['inheritref'])
+            inherit_node = node['inheritnode']
+            if node['inheritref'] not in inherits:
+                if app.config.inheritance_debug:
+                    sys.stderr.write("DEBUG: ref '%s' found in inheritref "
+                            "directive doesn't have any inheritance.\n"
+                                    % node['inheritref'])
+                continue
+            apply_inheritance(app, [node, inherit_node], node['inheritref'])
         return
     # no autoreferences
     for node in doctree.traverse():
@@ -315,7 +387,35 @@ def apply_inheritance(app, node_list, inheritref):
         if position == u'after':
             node_list[-1].replace_self([node_list[-1]] + inherit_nodes)
         elif position == u'before':
-            node_list[0].replace_self(inherit_nodes + node_list[0])
+            node_list[0].replace_self(inherit_nodes + [node_list[0]])
+        elif position == u'inside':
+            if inherit_vals['nodetype'] == 'toctree':
+                if app.config.verbose:
+                    sys.stderr.write("inherit_nodes of 'inside' 'toctree': "
+                            "%s\nnode_list: %s\n" % (inherit_nodes[0],
+                                    node_list[1]))
+                    sys.stderr.write("inherit_nodes['entries'] of 'inside' "
+                            "'toctree': %s\nnode_list['entries']: %s\n" % (
+                                    inherit_nodes[0]['entries'],
+                                    node_list[1]['entries']))
+                node_list[1]['entries'].extend(inherit_nodes[0]['entries'])
+                node_list[1]['includefiles'].extend(
+                        inherit_nodes[0]['includefiles'])
+            elif inherit_vals['nodetype'] == 'bullet_list':
+                if app.config.verbose:
+                    sys.stderr.write("inherit_nodes of 'inside' 'bullet_list':"
+                            " %s\nnode_list: %s\n" % (inherit_nodes[0],
+                                    node_list[1]))
+                    sys.stderr.write("inherit_nodes.children of 'inside' "
+                            "'bullet_list': %s\nnode_list.children: %s\n"
+                                    % (inherit_nodes[0].children,
+                                            node_list[1].children))
+                node_list[1].extend(inherit_nodes[0].children)
+                if len(inherit_nodes) > 1:
+                    node_list[-1].replace_self([node_list[-1]] +
+                            inherit_nodes[1:])
+            else:
+                node_list[-1].extend(inherit_nodes)
         inherit_vals['replaced'] += 1
     return
 
@@ -398,10 +498,10 @@ def report_warnings(app, exception):
 
 def setup(app):
     app.add_config_value('inheritance_plaintext', True, 'env')
-    app.add_config_value('inheritance_pattern', re.compile(r'#:(.|[^#]+)#'), 
+    app.add_config_value('inheritance_pattern', re.compile(r'^#\:(.|[^#]+)#$'), 
             'env')
     app.add_config_value('inheritance_reference_pattern',
-            re.compile(r'(?P<prefix>[a-zA-Z/_]+):(?P<nodetype>[a-z]+):'
+            re.compile(r'(?P<prefix>[a-zA-Z/_]+):(?P<nodetype>[a-z_]+):'
                     '(?P<identifier>[a-zA-Z]+)'), 'env')
     app.add_config_value('inheritance_modules', [], 'env')
     app.add_config_value('inheritance_autoreferences', False, 'env')
